# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'maxent_bosonic_ui.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(759, 629)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.real_freq_frame = QtWidgets.QFrame(self.centralwidget)
        self.real_freq_frame.setGeometry(QtCore.QRect(10, 10, 231, 171))
        self.real_freq_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.real_freq_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.real_freq_frame.setObjectName("real_freq_frame")
        self.label = QtWidgets.QLabel(self.real_freq_frame)
        self.label.setGeometry(QtCore.QRect(10, 10, 141, 17))
        self.label.setObjectName("label")
        self.label_3 = QtWidgets.QLabel(self.real_freq_frame)
        self.label_3.setGeometry(QtCore.QRect(10, 70, 31, 17))
        self.label_3.setObjectName("label_3")
        self.label_4 = QtWidgets.QLabel(self.real_freq_frame)
        self.label_4.setGeometry(QtCore.QRect(10, 110, 21, 17))
        self.label_4.setObjectName("label_4")
        self.grid_type_combo = QtWidgets.QComboBox(self.real_freq_frame)
        self.grid_type_combo.setGeometry(QtCore.QRect(10, 40, 201, 25))
        self.grid_type_combo.setObjectName("grid_type_combo")
        self.grid_type_combo.addItem("")
        self.grid_type_combo.addItem("")
        self.max_real_freq = QtWidgets.QLineEdit(self.real_freq_frame)
        self.max_real_freq.setGeometry(QtCore.QRect(40, 70, 41, 25))
        self.max_real_freq.setObjectName("max_real_freq")
        self.num_real_freq = QtWidgets.QLineEdit(self.real_freq_frame)
        self.num_real_freq.setGeometry(QtCore.QRect(40, 110, 41, 25))
        self.num_real_freq.setObjectName("num_real_freq")
        self.gen_real_grid_button = QtWidgets.QPushButton(self.real_freq_frame)
        self.gen_real_grid_button.setGeometry(QtCore.QRect(90, 110, 71, 25))
        self.gen_real_grid_button.setObjectName("gen_real_grid_button")
        self.input_data_tabs = QtWidgets.QTabWidget(self.centralwidget)
        self.input_data_tabs.setGeometry(QtCore.QRect(250, 10, 501, 171))
        self.input_data_tabs.setObjectName("input_data_tabs")
        self.text_tab = QtWidgets.QWidget()
        self.text_tab.setObjectName("text_tab")
        self.label_14 = QtWidgets.QLabel(self.text_tab)
        self.label_14.setGeometry(QtCore.QRect(10, 10, 221, 17))
        self.label_14.setObjectName("label_14")
        self.label_16 = QtWidgets.QLabel(self.text_tab)
        self.label_16.setGeometry(QtCore.QRect(10, 40, 67, 17))
        self.label_16.setObjectName("label_16")
        self.inp_file_name_2 = QtWidgets.QLineEdit(self.text_tab)
        self.inp_file_name_2.setGeometry(QtCore.QRect(90, 40, 361, 25))
        self.inp_file_name_2.setObjectName("inp_file_name_2")
        self.select_file_button_2 = QtWidgets.QToolButton(self.text_tab)
        self.select_file_button_2.setGeometry(QtCore.QRect(460, 40, 26, 24))
        self.select_file_button_2.setObjectName("select_file_button_2")
        self.load_data_button_2 = QtWidgets.QPushButton(self.text_tab)
        self.load_data_button_2.setGeometry(QtCore.QRect(300, 110, 89, 25))
        self.load_data_button_2.setObjectName("load_data_button_2")
        self.show_data_button_2 = QtWidgets.QPushButton(self.text_tab)
        self.show_data_button_2.setGeometry(QtCore.QRect(400, 110, 89, 25))
        self.show_data_button_2.setObjectName("show_data_button_2")
        self.label_17 = QtWidgets.QLabel(self.text_tab)
        self.label_17.setGeometry(QtCore.QRect(10, 80, 31, 17))
        self.label_17.setObjectName("label_17")
        self.n_skip = QtWidgets.QLineEdit(self.text_tab)
        self.n_skip.setGeometry(QtCore.QRect(50, 80, 31, 25))
        self.n_skip.setObjectName("n_skip")
        self.label_18 = QtWidgets.QLabel(self.text_tab)
        self.label_18.setGeometry(QtCore.QRect(90, 80, 151, 17))
        self.label_18.setObjectName("label_18")
        self.label_19 = QtWidgets.QLabel(self.text_tab)
        self.label_19.setGeometry(QtCore.QRect(10, 110, 31, 17))
        self.label_19.setObjectName("label_19")
        self.num_mats_freq_text = QtWidgets.QLineEdit(self.text_tab)
        self.num_mats_freq_text.setGeometry(QtCore.QRect(50, 110, 31, 25))
        self.num_mats_freq_text.setObjectName("num_mats_freq_text")
        self.label_20 = QtWidgets.QLabel(self.text_tab)
        self.label_20.setGeometry(QtCore.QRect(90, 110, 161, 17))
        self.label_20.setObjectName("label_20")
        self.input_data_tabs.addTab(self.text_tab, "")
        self.continuation_frame = QtWidgets.QFrame(self.centralwidget)
        self.continuation_frame.setGeometry(QtCore.QRect(10, 190, 741, 391))
        self.continuation_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.continuation_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.continuation_frame.setObjectName("continuation_frame")
        self.doit_button = QtWidgets.QPushButton(self.continuation_frame)
        self.doit_button.setGeometry(QtCore.QRect(590, 20, 131, 41))
        self.doit_button.setObjectName("doit_button")
        self.blur_width = QtWidgets.QLineEdit(self.continuation_frame)
        self.blur_width.setGeometry(QtCore.QRect(80, 40, 113, 25))
        self.blur_width.setObjectName("blur_width")
        self.label_11 = QtWidgets.QLabel(self.continuation_frame)
        self.label_11.setGeometry(QtCore.QRect(30, 40, 51, 17))
        self.label_11.setObjectName("label_11")
        self.text_output = QtWidgets.QTextEdit(self.continuation_frame)
        self.text_output.setGeometry(QtCore.QRect(30, 80, 691, 231))
        self.text_output.setObjectName("text_output")
        self.save_button = QtWidgets.QPushButton(self.continuation_frame)
        self.save_button.setGeometry(QtCore.QRect(630, 360, 89, 25))
        self.save_button.setObjectName("save_button")
        self.output_directory_button = QtWidgets.QToolButton(self.continuation_frame)
        self.output_directory_button.setGeometry(QtCore.QRect(120, 360, 26, 24))
        self.output_directory_button.setObjectName("output_directory_button")
        self.out_file_name = QtWidgets.QLineEdit(self.continuation_frame)
        self.out_file_name.setGeometry(QtCore.QRect(160, 360, 451, 25))
        self.out_file_name.setObjectName("out_file_name")
        self.label_12 = QtWidgets.QLabel(self.continuation_frame)
        self.label_12.setGeometry(QtCore.QRect(30, 360, 91, 17))
        self.label_12.setObjectName("label_12")
        self.n_interpolation = QtWidgets.QLineEdit(self.continuation_frame)
        self.n_interpolation.setGeometry(QtCore.QRect(200, 320, 41, 25))
        self.n_interpolation.setObjectName("n_interpolation")
        self.label_13 = QtWidgets.QLabel(self.continuation_frame)
        self.label_13.setGeometry(QtCore.QRect(250, 320, 201, 17))
        self.label_13.setObjectName("label_13")
        self.preblur_checkbox = QtWidgets.QCheckBox(self.continuation_frame)
        self.preblur_checkbox.setGeometry(QtCore.QRect(30, 10, 92, 23))
        self.preblur_checkbox.setObjectName("preblur_checkbox")
        self.interpolate_checkbox = QtWidgets.QCheckBox(self.continuation_frame)
        self.interpolate_checkbox.setGeometry(QtCore.QRect(30, 320, 171, 23))
        self.interpolate_checkbox.setObjectName("interpolate_checkbox")
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 759, 22))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        self.input_data_tabs.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.real_freq_frame.setToolTip(_translate("MainWindow", "configure real-frequency grid"))
        self.real_freq_frame.setWhatsThis(_translate("MainWindow", "real-frequency grid"))
        self.label.setText(_translate("MainWindow", "Real-frequency grid"))
        self.label_3.setText(_translate("MainWindow", "max"))
        self.label_4.setText(_translate("MainWindow", "n"))
        self.grid_type_combo.setToolTip(_translate("MainWindow", "equispaced or centered grid (denser around Fermi energy)"))
        self.grid_type_combo.setItemText(0, _translate("MainWindow", "equispaced positive"))
        self.grid_type_combo.setItemText(1, _translate("MainWindow", "centered positive"))
        self.max_real_freq.setToolTip(_translate("MainWindow", "upper border of real-frequency grid. (lower border is set symmetrically)"))
        self.max_real_freq.setText(_translate("MainWindow", "20"))
        self.num_real_freq.setToolTip(_translate("MainWindow", "number frequencies on real axis; should be an odd number"))
        self.num_real_freq.setText(_translate("MainWindow", "401"))
        self.gen_real_grid_button.setText(_translate("MainWindow", "Generate"))
        self.label_14.setText(_translate("MainWindow", "Load susceptibility from text file"))
        self.label_16.setText(_translate("MainWindow", "file name"))
        self.inp_file_name_2.setToolTip(_translate("MainWindow", "file path and name of a w2dynamics output file"))
        self.select_file_button_2.setToolTip(_translate("MainWindow", "choose an input file"))
        self.select_file_button_2.setText(_translate("MainWindow", "..."))
        self.load_data_button_2.setText(_translate("MainWindow", "Load data"))
        self.show_data_button_2.setToolTip(_translate("MainWindow", "click this if you want to plot the data after loading"))
        self.show_data_button_2.setText(_translate("MainWindow", "Show data"))
        self.label_17.setText(_translate("MainWindow", "Skip"))
        self.label_18.setText(_translate("MainWindow", "lines at the beginning"))
        self.label_19.setText(_translate("MainWindow", "Use"))
        self.label_20.setText(_translate("MainWindow", "Matsubara frequencies"))
        self.input_data_tabs.setTabText(self.input_data_tabs.indexOf(self.text_tab), _translate("MainWindow", "text file"))
        self.doit_button.setToolTip(_translate("MainWindow", "perform the analytical continuation"))
        self.doit_button.setText(_translate("MainWindow", "Do it!"))
        self.blur_width.setToolTip(_translate("MainWindow", "set the blur width here"))
        self.blur_width.setText(_translate("MainWindow", "0.1"))
        self.label_11.setText(_translate("MainWindow", "Width"))
        self.text_output.setToolTip(_translate("MainWindow", "in this field some output will be shown"))
        self.save_button.setToolTip(_translate("MainWindow", "click this button to save the output"))
        self.save_button.setText(_translate("MainWindow", "Save"))
        self.output_directory_button.setToolTip(_translate("MainWindow", "Choose a directory, where you want to save the output"))
        self.output_directory_button.setText(_translate("MainWindow", "..."))
        self.out_file_name.setToolTip(_translate("MainWindow", "type full output name here (including path)"))
        self.label_12.setText(_translate("MainWindow", "Output file:"))
        self.n_interpolation.setToolTip(_translate("MainWindow", "number of regularly spaced grid points for interpolation"))
        self.n_interpolation.setText(_translate("MainWindow", "0"))
        self.label_13.setText(_translate("MainWindow", "regularly spaced grid points"))
        self.preblur_checkbox.setToolTip(_translate("MainWindow", "check this if you want to use preblur"))
        self.preblur_checkbox.setText(_translate("MainWindow", "Preblur"))
        self.interpolate_checkbox.setToolTip(_translate("MainWindow", "check this for interpolating output to regular grid"))
        self.interpolate_checkbox.setText(_translate("MainWindow", "Interpolate output to"))

